use std::time::Instant;

use futures_util::{AsyncReadExt as _, AsyncWriteExt as _};
use mongodb::{options::GridFsBucketOptions, Client};
use rand::RngCore as _;
use thiserror::Error;

#[derive(Debug, Error)]
enum Error {
    #[error("io error: {0}")]
    IoError(#[from] std::io::Error),
    #[error("rand error: {0}")]
    RandError(#[from] rand::Error),
    #[error("mongodb error: {0}")]
    MongoDBError(#[from] mongodb::error::Error),
}

type Result<T> = std::result::Result<T, Error>;

#[tokio::main]
async fn main() -> Result<()> {
    println!("rust mongodb: 100 x 1MiB");

    let mut data = [0u8; 1000000];
    rand::thread_rng().try_fill_bytes(&mut data)?;

    let client = Client::with_uri_str("mongodb://127.0.0.1")
        .await?
        .database("test_rust_mongodb");

    let bucket = client.gridfs_bucket(GridFsBucketOptions::default());

    let now = Instant::now();
    let mut stream = bucket.open_upload_stream("test").await?;
    for _ in 0..100 {
        stream.write_all(&data).await?;
    }
    stream.close().await?;
    let elapsed = now.elapsed();
    println!("    WRITE: {}ms", elapsed.as_millis());

    {
        let now = Instant::now();

        let mut stream = bucket.open_download_stream_by_name("test").await?;
        let mut buf = Vec::new();
        stream.read_to_end(&mut buf).await?;

        let elapsed = now.elapsed();
        println!("    READ FULL: {}ms", elapsed.as_millis());
    }

    {
        let now = Instant::now();

        let mut stream = bucket
            .open_download_stream_by_name("test")
            .start(1055)
            .end(1481000)
            .await?;
        let mut buf = Vec::new();
        stream.read_to_end(&mut buf).await?;
        let elapsed = now.elapsed();
        println!("    READ PARTIAL: {}ms", elapsed.as_millis());

        assert_eq!(data[1055..], buf[..998945]);
        assert_eq!(data[..481000], buf[998945..]);
        println!("{:?}", &data[1055..1065]);
        println!("{:?}", &buf[0..10]);
        println!("{:?}", &data[0..10]);
        println!("{:?}", &buf[998945..998955]);
    }

    Ok(())
}
